'
' PBDRIVE include file - Public Domain by Borje Hagsten, January 2002.
' Totally self-containing via container window.
'
' LOG:
' Aug 15, 2003: Fixed a call to GetLogicalDriveStrings in DrivesToCombo
'               where it were wrongly given %MAX_PATH instead of real
'               length, which could cause it to fail in some situations.
' Jan 14, 2003: Changed handles to DWORD and now use GetSystemMetrics
'               to ensure item height = small icon if big font is set.
'               It now uses GetWindowLong(hWnd, %GWL_HINSTANCE) to
'               get proper instance handle at creation.
'
%DCM_SETDRVDIRCTL    = %WM_USER + 4000 ' tell what DIR list control is attached to drive combo
%DCM_SETDRVSEL3DFACE = %WM_USER + 4001 ' wParam sets custom sel color, lParam decides redraw
'--------------------------------------------------------------------
TYPE PBDRIVECTLDATA
  hDir       AS DWORD
  oldProc    AS DWORD
  sel3Dface  AS LONG
  curPath    AS ASCIIZ * %MAX_PATH
END TYPE
'--------------------------------------------------------------------
DECLARE SUB DrivesToCombo(BYVAL hWnd AS DWORD, BYVAL Path AS STRING)
DECLARE FUNCTION PBDriveProc(BYVAL hWnd AS DWORD, BYVAL wMsg AS DWORD, _
                             BYVAL wParam AS LONG, BYVAL lParam AS LONG) AS LONG
DECLARE FUNCTION PBDrivePath (BYVAL hDlg AS DWORD, BYVAL DriveId AS DWORD, _
                              BYVAL sPBDrivePath AS STRING) AS STRING

'
' Create Color Combobox list
'
FUNCTION CreatePBDriveCB ALIAS "CreatePBDriveCB" _
                        (BYVAL hWnd AS DWORD, BYVAL CtrlId AS LONG, _
                         BYVAL vLeft AS LONG, BYVAL vTop AS LONG, _
                         BYVAL vWidth AS LONG, BYVAL vHeight AS LONG, _
                         BYVAL DDTstyle AS LONG, BYVAL cDir AS STRING) EXPORT AS DWORD

  LOCAL hParent AS DWORD, hCombo AS DWORD, cbFont AS DWORD, rc AS RECT
  LOCAL wc AS WNDCLASSEX, szClassName AS ASCIIZ * 20, pdd AS PBDRIVECTLDATA PTR

  'REGISTER A CUSTOM CONTROL CLASS FOR CONTAINER WINDOW
  szClassName      = "PBDRIVEX"
  wc.cbSize        = LEN(wc)
  wc.lpfnWndProc   = CODEPTR(PBDriveMainProc)
  wc.cbWndExtra    = 4
  wc.hInstance     = GetWindowLong(hWnd, %GWL_HINSTANCE)
  wc.lpszClassName = VARPTR(szClassName)

  RegisterClassEx wc

  IF DDTstyle THEN
     DIALOG UNITS hWnd, vLeft, vTop TO PIXELS vLeft, vTop
     DIALOG UNITS hWnd, vWidth, vHeight TO PIXELS vWidth, vHeight
  END IF

  'CREATE CONTAINER WINDOW, for redrawing operations in WM_DRAWITEM
  hParent = CreateWindowEx(0, "PBDRIVEX", BYVAL 0, %WS_VISIBLE OR %WS_CHILD OR %WS_TABSTOP, _
                           vLeft, vTop, vWidth, vHeight, _
                           hwnd, CtrlId, wc.hInstance, BYVAL 0)

  IF hParent THEN 'CREATE OWNERDRAWN COMBOBOX
     hCombo = CreateWindowEx(%WS_EX_CLIENTEDGE, "COMBOBOX", BYVAL 0, %WS_VISIBLE OR _
                             %WS_CHILD OR %WS_VISIBLE OR %CBS_OWNERDRAWFIXED OR %CBS_HASSTRINGS OR _
                             %CBS_DROPDOWNLIST OR %WS_TABSTOP OR %CBS_DISABLENOSCROLL OR %WS_VSCROLL, _
                             0, 0, vWidth, vHeight, _
                             hParent, CtrlId, wc.hInstance, BYVAL 0)
  ELSE
     EXIT FUNCTION 'return zero on failure
  END IF

  IF hCombo THEN
     pdd = HeapAlloc(GetProcessHeap(), %HEAP_ZERO_MEMORY, SIZEOF(@pdd))
     IF pdd THEN 'store pointer to memory for UDT in parent's cbWndExtra
        SetWindowLong hParent, 0, pdd
      ELSE
        EXIT FUNCTION
     END IF

     'set standard Windows font
     cbFont = GetStockObject(%ANSI_VAR_FONT)
     IF cbFont THEN SendMessage hCombo, %WM_SETFONT, cbFont, 0

     'ensure item height is at least small icon (caption icon) height
     SendMessage hCombo, %CB_SETITEMHEIGHT, 0, GetSystemMetrics(%SM_CYSMICON) + 3

     GetWindowRect hCombo, rc 'must resize parent to actual combo size
     SetWindowPos hParent, 0, 0, 0, rc.nRight - rc.nLeft, _
                  rc.nBottom - rc.nTop, %SWP_NOMOVE OR %SWP_NOZORDER

     'Subclass control and store pointer to old procedure in parent's cbWndExtra
     @pdd.oldProc = SetWindowLong(hCombo, %GWL_WNDPROC, BYVAL CODEPTR(PBDriveProc))

     IF cDir = "" THEN cDir = CURDIR$
     DrivesToCombo hCombo, cDir
  END IF

  FUNCTION = hCombo 'return the handle of combobox
END FUNCTION

'
' Container procedure
'
FUNCTION PBDriveMainProc(BYVAL hWnd   AS DWORD, BYVAL wMsg   AS DWORD,_
                         BYVAL wParam AS LONG, BYVAL lParam AS LONG) AS LONG
  LOCAL pdd AS PBDRIVECTLDATA PTR

  SELECT CASE wMsg
     CASE %WM_CREATE

     CASE %WM_SETFOCUS
        SetFocus GetWindow(hWnd, %GW_CHILD)  'Move focus to next control
        FUNCTION = 0 : EXIT FUNCTION

     CASE %WM_DESTROY 'Un-subclass listbox when it's detroyed
        pdd = GetWindowLong(hWnd, 0)    'Get stored data
        IF @pdd.oldProc THEN SetWindowLong GetWindow(hWnd, %GW_CHILD), %GWL_WNDPROC, @pdd.oldProc
        HeapFree GetProcessHeap(), 0, BYVAL pdd  'free memory
        FUNCTION = 0 : EXIT FUNCTION

     CASE %WM_DRAWITEM 'pass on to control
        SendMessage GetDlgItem(hWnd, wParam), wMsg, wParam, lParam
        FUNCTION = %TRUE : EXIT FUNCTION

     CASE %WM_COMMAND  'pass these on to container's parent
        IF HIWRD(wParam) = %CBN_SELENDOK THEN
           pdd = GetWindowLong(hWnd, 0)    'Get stored data
           '-----------------------------------------
           'avoid re-select of last selected item if mouse is clicked outside list
           LOCAL hDrive AS DWORD, itemCount AS LONG, itemHeight AS LONG, visItems AS LONG, _
                 rc AS RECT, rc2 AS RECT, pt AS POINTAPI
           hDrive = GetWindow(hWnd, %GW_CHILD)
           GetCursorPos pt
           GetWindowRect hDrive, rc2
           itemCount  = SendMessage(hDrive, %CB_GETCOUNT, 0, 0)
           itemHeight = SendMessage(hDrive, %CB_GETITEMHEIGHT, 0, 0)
           SendMessage hDrive, %CB_GETDROPPEDCONTROLRECT, 0, VARPTR(rc)

           rc.nTop    = rc.nTop + rc2.nBottom - rc2.nTop
           visItems   = MIN&(itemCount, (rc.nBottom - rc.nTop) \ itemHeight)
           rc.nBottom = rc.nTop + visItems * itemHeight

           IF ((GetKeyState(%VK_RETURN) AND &H8000) = 0 AND PtInRect(rc, pt.x, pt.y) = 0) OR _
                 (GetKeyState(%VK_ESCAPE) AND &H8000) THEN
              EXIT FUNCTION
           END IF
           '-----------------------------------------

           LOCAL lRs AS LONG, curPath AS STRING, newPath AS STRING, _
                 zTxt AS ASCIIZ * 300, zp AS ASCIIZ PTR
           zp = SendMessage(@pdd.hDir, %WM_USER + 5010, 0, 0) '%DCM_GETDIRPATH
           curPath = @zp
           newPath = PBDrivePath(GetParent(hWnd), LOWRD(wParam), "")

           SetCurrentDirectory BYVAL STRPTR(newPath)
           lRs = GetLastError
           IF lRs THEN
              FormatMessage %FORMAT_MESSAGE_FROM_SYSTEM, BYVAL 0, lRs, _
                             MAKLNG(%LANG_NEUTRAL, %SUBLANG_DEFAULT), _
                             zTxt, SIZEOF(zTxt), BYVAL 0
              MessageBox GetParent(hWnd), zTxt, "System message", %MB_OK OR %MB_ICONEXCLAMATION
              IF LEN(curPath) THEN                                'new path gave error, reset to old path again
                 PBDrivePath GetParent(hWnd), LOWRD(wParam), curPath
                 EXIT FUNCTION
              END IF
           END IF

           IF @pdd.hDir THEN SendMessage @pdd.hDir, %WM_USER + 5004, 0, BYVAL STRPTR(newPath) '%DCM_SETDIRS
        END IF
           
        SendMessage GetParent(hWnd), wMsg, wParam, lParam
        FUNCTION = 0 : EXIT FUNCTION
  END SELECT

  FUNCTION = DefWindowProc(hWnd, wMsg, wParam, lParam) 'default message processing
END FUNCTION

'
' Subclassed Combobox procedure
'
FUNCTION PBDriveProc(BYVAL hWnd AS DWORD, BYVAL wMsg AS DWORD, _
                     BYVAL wParam AS LONG, BYVAL lParam AS LONG) AS LONG
  LOCAL pdd AS PBDRIVECTLDATA PTR
  pdd = GetWindowLong(GetParent(hWnd), 0) 'Get stored data

  SELECT CASE wMsg
     CASE %DCM_SETDRVDIRCTL 'set associated DIR control
        @pdd.hDir = wParam

     CASE %DCM_SETDRVSEL3DFACE ' set control to use custom sel color
        @pdd.sel3Dface = wParam
        IF lParam THEN
           InvalidateRect hWnd, BYVAL %NULL, 0
           UpdateWindow hWnd
        END IF

     CASE %WM_DRAWITEM
        LOCAL lpdis AS DRAWITEMSTRUCT PTR
         lpdis = lParam
        IF @lpdis.itemID = &HFFFFFFFF& THEN EXIT FUNCTION
 
        SELECT CASE @lpdis.itemAction
           CASE %ODA_DRAWENTIRE, %ODA_SELECT
              LOCAL y AS LONG, lMarg AS LONG, hIcon AS DWORD, selCol AS LONG, zTxt AS ASCIIZ * 260
              LOCAL TextDrawFlag AS DWORD, IconDrawFlag AS DWORD, tm AS TEXTMETRIC
              TextDrawFlag = &H2  ' DST_PREFIXTEXT
              IconDrawFlag = &H3  ' DST_ICON
              selCol = GetWindowLong(GetParent(hWnd), 8) 'is control set for custom sel color?
 
              'CLEAR BACKGROUND
              IF IsWindowEnabled(hWnd) THEN                           'ENABLED CONTROL
                 IF (@lpdis.itemState AND %ODS_SELECTED) = 0 OR _     'if not selected
                    (@lpdis.itemState AND &H1000) THEN                'or is %ODS_COMBOBOXEDIT part
                    FillRect @lpdis.hDC, @lpdis.rcItem, GetSysColorBrush(%COLOR_WINDOW) 'clear background
                    SetBkColor @lpdis.hDC, GetSysColor(%COLOR_WINDOW)        'text background
                    SetTextColor @lpdis.hDC, GetSysColor(%COLOR_WINDOWTEXT)  'text color
                 ELSE
                    IF @pdd.sel3Dface THEN 'control is set to use custom sel color
                       FillRect @lpdis.hDC, @lpdis.rcItem, GetSysColorBrush(%COLOR_3DFACE) 'clear background
                       SetBkColor @lpdis.hDC, GetSysColor(%COLOR_3DFACE)     'text background
                       SetTextColor @lpdis.hDC, GetSysColor(%COLOR_BTNTEXT)  'text color
                    ELSE
                       FillRect @lpdis.hDC, @lpdis.rcItem, GetSysColorBrush(%COLOR_HIGHLIGHT) 'clear background
                       SetBkColor @lpdis.hDC, GetSysColor(%COLOR_HIGHLIGHT)        'text background
                       SetTextColor @lpdis.hDC, GetSysColor(%COLOR_HIGHLIGHTTEXT)  'text color
                    END IF
                 END IF
 
              ELSE                                                        'DISABLED CONTROL
                 FillRect @lpdis.hDC, @lpdis.rcItem, GetSysColorBrush(%COLOR_3DFACE) 'clear background
                 SetBkColor @lpdis.hDC, GetSysColor(%COLOR_3DFACE)        'text background
                 SetTextColor @lpdis.hDC, GetSysColor(%COLOR_WINDOWTEXT)  'text color
                 TextDrawFlag = TextDrawFlag OR &H20 ' DSS_DISABLED
                 IconDrawFlag = IconDrawFlag OR &H20 ' DSS_DISABLED
              END IF
 
              IF (@lpdis.itemState AND &H1000) THEN 'if to draw %ODS_COMBOBOXEDIT (= &H1000)
                 lMarg = 6
              ELSE
                 lMarg = 4
              END IF

              'CALCULATE CENTERED Y POS
              GetTextMetrics @lpdis.hDC, tm                                      'get font size
              y = (@lpdis.rcItem.nBottom + @lpdis.rcItem.nTop - tm.tmHeight) / 2 'center text

              'DRAW TEXT (Note - DrawState does not work with NT 3.5..)
              SendMessage hWnd, %CB_GETLBTEXT, @lpdis.itemID, VARPTR(zTxt) 'text used to retrieve data
              DrawState @lpdis.hDC, 0, 0, BYVAL VARPTR(zTxt), LEN(zTxt), _
                             lMarg + GetSystemMetrics(%SM_CYCAPTION) + 1, y, _
                             @lpdis.rcItem.nRight, @lpdis.rcItem.nBottom, TextDrawFlag
 
              'DRAW ASSOCIATED ICON (brush in hbr member is ignored for all states except DSS_MONO)
              hIcon = SendMessage(hWnd, %CB_GETITEMDATA, @lpdis.itemID, 0)
              IF hIcon THEN
                 DrawState @lpdis.hDC, 0, 0, hIcon, 0, _
                           lMarg, @lpdis.rcItem.ntop + 1, 0, 0, IconDrawFlag
              END IF
 
              'SELECTED ITEM
              IF (@lpdis.itemState AND %ODS_SELECTED) THEN  'if selected
                 DrawFocusRect @lpdis.hDC, @lpdis.rcItem    'draw a focus rect around all
              END IF

        END SELECT
        FUNCTION = %TRUE : EXIT FUNCTION

  END SELECT

  FUNCTION = CallWindowProc(@pdd.oldProc, hWnd, wMsg, wParam, lParam) 'process other messages
END FUNCTION

'
' Populate Combobox with available drives
'
SUB DrivesToCombo(BYVAL hWnd AS DWORD, BYVAL Path AS STRING)
  LOCAL ii AS LONG, curdrv AS LONG, seldrv AS LONG, _
        ppos AS LONG, Root AS STRING, dwRes AS DWORD
  LOCAL lpshfi AS SHFILEINFO

  curdrv = ASC(UCASE$(Path))                                'drive to select
  dwRes = GetLogicalDriveStrings(0, BYVAL %NULL)            'get needed string length
  Root = SPACE$(dwRes + 2)                                  'allocate enough big buffer
  dwRes = GetLogicalDriveStrings(dwRes, BYVAL STRPTR(Root)) 'get local drives
 
  IF LEN(TRIM$(Root)) THEN
     SendMessage hWnd, %CB_RESETCONTENT, 0, 0  'clear CB, just in case..
 
     FOR ii = 1 TO TALLY(Root, CHR$(0)) - 1    'Parse string and add drives to combo
        Path = PARSE$(Root, CHR$(0), ii)
        IF LEN(Path) THEN
           SHGetFileInfo BYVAL STRPTR(Path), 0, lpshfi, LEN(lpshfi), _
                         %SHGFI_ICON OR %SHGFI_SMALLICON OR %SHGFI_DISPLAYNAME

           Path = UCASE$(LEFT$(Path, 2))
           ppos = INSTR(lpshfi.szDisplayName, "(")
           IF ppos THEN
              lpshfi.szDisplayName = TRIM$(LEFT$(lpshfi.szDisplayName, ppos - 1))
              IF LEN(lpshfi.szDisplayName) THEN
                 Path = Path + "  (" + lpshfi.szDisplayName + ")"
              END IF
           END IF
           SendMessage hWnd, %CB_ADDSTRING, 0, BYVAL STRPTR(Path)
           SendMessage hWnd, %CB_SETITEMDATA, ii - 1, lpshfi.hIcon
           IF curdrv = ASC(Path) THEN seldrv = ii - 1
        END IF
     NEXT

     SendMessage hWnd, %CB_SETCURSEL, seldrv, 0
  END IF
END SUB

'
' Return or set drive control's path
'
FUNCTION PBDrivePath (BYVAL hDlg AS DWORD, BYVAL DriveId AS DWORD, BYVAL sPBDrivePath AS STRING) AS STRING
  LOCAL hWnd AS DWORD, cLen AS LONG, cSel AS LONG
  hWnd = GetDlgItem(GetDlgItem(hDlg, DriveId), DriveId) 'list is child of container

  IF LEN(sPBDrivePath) THEN 'set path
     sPBDrivePath = UCASE$(LEFT$(sPBDrivePath, 1))
     cSel = SendMessage(hWnd, %CB_SELECTSTRING, -1, BYVAL STRPTR(sPBDrivePath))
  END IF

  cSel = SendMessage(hWnd, %CB_GETCURSEL, 0, 0)
  IF cSel <> %CB_ERR THEN
     cLen = SendMessage(hWnd, %CB_GETLBTEXTLEN, cSel, 0)
     sPBDrivePath = SPACE$(cLen)
     SendMessage hWnd, %CB_GETLBTEXT, cSel, STRPTR(sPBDrivePath)
  END IF

  FUNCTION = LEFT$(sPBDrivePath, 1) + ":\"
END FUNCTION
